package org.implab.gradle.containers;

import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

import org.gradle.api.DefaultTask;
import org.gradle.api.Plugin;
import org.gradle.api.Project;
import org.gradle.api.artifacts.Configuration;
import org.gradle.api.artifacts.Configuration.State;
import org.gradle.api.logging.Logger;
import org.gradle.api.tasks.Copy;
import org.gradle.api.tasks.Delete;
import org.implab.gradle.containers.cli.Utils;
import org.implab.gradle.containers.tasks.ComposeRm;
import org.implab.gradle.containers.tasks.ComposeStop;
import org.implab.gradle.containers.tasks.ComposeUp;
import org.implab.gradle.containers.tasks.WriteEnv;

public abstract class ComposePlugin implements Plugin<Project>, ProjectMixin {
    public static final String BUILD_GROUP = "build";

    public final String COMPOSE_IMAGES_CONFIGURATION = "composeImages";

    public final String COMPOSE_PROJECT_NAME = "COMPOSE_PROJECT_NAME";

    public final String COMPOSE_EXTENSION = "compose";

    public final String COMPOSE_UP_TASK = "up";

    public final String COMPOSE_STOP_TASK = "stop";

    public final String COMPOSE_RM_TASK = "rm";

    public final String CLEAN_TASK = "clean";

    public final String BUILD_TASK = "build";

    public final String PROCESS_RESOURCES_TASK = "processResources";

    public final String WRITE_ENV_TASK = "writeEnv";

    public final String COMPOSE_VAR = "composeVar";

    public final String ENV_FILE_NAME = ".env";

    public Logger getLogger() {
        return getProject().getLogger();
    }

    @Override
    public void apply(Project project) {
        var containerImages = configuration(COMPOSE_IMAGES_CONFIGURATION, Configurations.RESOLVABLE);

        // basic configuration, register extension
        plugin(ContainerBasePlugin.class);

        var composeExtension = extension(COMPOSE_EXTENSION, ComposeExtension.class);

        var cleanTask = task(CLEAN_TASK, Delete.class, t -> {
            t.delete(composeExtension.getContextDirectory());
        });

        // copy task from src/main
        var processResources = task(PROCESS_RESOURCES_TASK, Copy.class, t -> {
            t.mustRunAfter(cleanTask);
            t.from(projectDirectory().dir("src/main"));
            t.into(composeExtension.getContextDirectory());
        });

        // write .env
        var writeEnvTask = task(WRITE_ENV_TASK, WriteEnv.class, t -> {
            t.dependsOn(processResources, containerImages);
            t.getEnvFile().set(composeExtension.getContextDirectory().file(ENV_FILE_NAME));

            var group = project.getGroup();
            if (group != null && group.toString().length() > 0) {
                t.getEnvironment().put(COMPOSE_PROJECT_NAME, group.toString());
            }

            t.getEnvironment().putAll(containerImages.map(this::extractComposeEnv));

        });

        var buildTask = task(BUILD_TASK, DefaultTask.class, t -> {
            t.setGroup(BUILD_GROUP);
            t.dependsOn(writeEnvTask);
        });

        var stopTask = task(COMPOSE_STOP_TASK, ComposeStop.class, t -> {
            // stop must run after build
            t.mustRunAfter(buildTask);

        });

        var rmTask = task(COMPOSE_RM_TASK, ComposeRm.class, t -> {
            // rm must run after build and stop
            t.mustRunAfter(buildTask, stopTask);

        });

        task(COMPOSE_UP_TASK, ComposeUp.class, t -> {
            t.dependsOn(buildTask);
            // up must run after stop and rm
            t.mustRunAfter(stopTask, rmTask);

        });
    }

    /**
     * Processed the configurations, extracts composeVar extra property from
     * each dependency in this configuration and adds a value to the resulting
     * map. The values in the nap will contain image tags.
     */
    private Map<String, String> extractComposeEnv(Configuration config) {
        if (config.getState() != State.UNRESOLVED) {
            getLogger().error("extractComposeEnv: The configuration {} isn't resolved.", config.getName());
            throw new IllegalStateException("The specified configuration isn't resolved");
        }

        getLogger().info("extractComposeEnv {}", config.getName());

        var map = new HashMap<String, String>();

        for (var dependency : config.getDependencies()) {
            // get extra composeVar if present
            extra(dependency, COMPOSE_VAR, String.class).optional().ifPresent(varName -> {
                // if we have a composeVar extra attribute on this dependency

                // get files for the dependency
                var files = config.files(dependency);
                if (files.size() == 1) {
                    // should bw exactly 1 file
                    var file = files.stream().findAny().get();
                    getLogger().info("Processing {}: {}", dependency, file);

                    try {
                        // try read imageRef
                        Utils.readImageRef(file).getTag()
                                .ifPresentOrElse(
                                        tag -> map.put(varName, tag),
                                        () -> getLogger().error("ImageRef doesn't have a tag: {}", file));

                    } catch (IOException e) {
                        getLogger().error("Failed to read ImageRef {}: {}", file, e);
                    }

                } else {
                    getLogger().warn("Dependency {} must have exactly 1 file", dependency);
                }
            });
        }

        return map;
    }

}
