package org.implab.gradle.containers.cli;

import java.io.File;
import java.io.IOException;
import java.util.List;
import java.util.Set;
import java.util.concurrent.ExecutionException;

public abstract class DockerTraits {

    private final String BUILD_COMMAND = "build";
    private final String PUSH_COMMAND = "push";
    private final String RUN_COMMAND = "run";
    private final String EXEC_COMMAND = "exec";
    private final String SAVE_COMMAND = "save";
    private final String INSPECT_COMMAND = "inspect";
    private final String IMAGE_COMMAND = "image";
    private final String TAG_COMMAND = "tag";
    private final String COMPOSE_COMMAND = "compose";
    private final String STOP_COMMAND = "stop";
    private final String START_COMMAND = "start";
    private final String CONTAINER_COMMAND = "container";
    private final String RM_COMMAND = "rm";

    protected abstract ProcessSpec builder(String... command);

    public ProcessSpec imageBuild(String imageName, File contextDirectory, List<String> options) {
        return builder(BUILD_COMMAND)
                .args(options)
                .args("-t", imageName, contextDirectory.getAbsolutePath());
    }

    public ProcessSpec imagePush(String image, List<String> options) {
        return builder(PUSH_COMMAND)
                .args(options)
                .args(image);
    }

    public ProcessSpec containerRun(String image, List<String> options, List<String> command) {
        return builder(RUN_COMMAND)
                .args(options)
                .args(image)
                .args(command);
    }

    public ProcessSpec containerExec(String containerId, List<String> options, List<String> command) {
        return builder(EXEC_COMMAND)
                .args(options)
                .args(containerId)
                .args(command);
    }

    public ProcessSpec containerStop(String containerId, List<String> options) {
        return builder(STOP_COMMAND, containerId).args(options);
    }

    public ProcessSpec containerStart(Set<String> containers, List<String> options) {
        return builder(START_COMMAND).args(options).args(containers);
    }

    public ProcessSpec containerRm(Set<String> containers, boolean removeVolumes, boolean force) {
        var spec = builder(RM_COMMAND);

        if(removeVolumes)
            spec.args("--volumes");
        if (force)
            spec.args("--force");

        spec.args(containers);

        return spec;
    }

    public ProcessSpec imageSave(Set<String> images, File output) {
        if (output.exists())
            output.delete();

        return builder(SAVE_COMMAND)
                .args("-o", output.getAbsolutePath())
                .args(images);
    }

    public ProcessSpec imageTag(String source, String target) {
        return builder(TAG_COMMAND, source, target);
    }

    public ProcessSpec imageExists(String imageId) throws InterruptedException, IOException, ExecutionException {
        return builder(IMAGE_COMMAND, INSPECT_COMMAND, "--format", "image-exists", imageId);
    }

    public ProcessSpec containerExists(String imageId) throws InterruptedException, IOException, ExecutionException {
        return builder(CONTAINER_COMMAND, INSPECT_COMMAND, "--format", "container-exists", imageId);
    }

    public ComposeTraits compose(File primaryCompose, Set<String> profiles) {
        return new ComposeTraits() {
            @Override
            public ProcessSpec compose(String command) {
                var spec = builder(COMPOSE_COMMAND, "--file", primaryCompose.getAbsolutePath());

                for (var profile : profiles)
                    spec.args("--profile", profile);
                return spec.args(command);
            }
        };
    }
}
