package org.implab.gradle.containers.cli;

import java.io.ByteArrayOutputStream;
import java.io.Closeable;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.Reader;
import java.util.Scanner;
import java.util.Set;
import java.util.concurrent.CompletableFuture;
import java.util.stream.Collectors;
import java.util.stream.StreamSupport;
import java.nio.file.Files;
import java.util.List;
import org.gradle.api.Action;
import org.gradle.api.GradleException;
import org.gradle.api.file.FileSystemLocation;
import org.gradle.internal.impldep.org.bouncycastle.util.Iterable;

import com.fasterxml.jackson.core.exc.StreamWriteException;
import com.fasterxml.jackson.databind.DatabindException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.SerializationFeature;
import com.fasterxml.jackson.datatype.jdk8.Jdk8Module;

import groovy.json.JsonGenerator;
import groovy.json.JsonOutput;
import groovy.json.JsonGenerator.Converter;
import groovy.lang.Closure;

public final class Utils {
    public static CompletableFuture<Void> redirectIO(final InputStream src, final Action<String> consumer) {
        return CompletableFuture.runAsync(() -> {
            try (Scanner sc = new Scanner(src)) {
                while (sc.hasNextLine()) {
                    consumer.execute(sc.nextLine());
                }
            }
        });
    }

    public static CompletableFuture<Void> redirectIO(final InputStream src, final File file) {
        return CompletableFuture.runAsync(() -> {
            try (OutputStream out = new FileOutputStream(file)) {
                src.transferTo(out);
            } catch (Exception e) {
                // silence!
            }
        });
    }

    public static CompletableFuture<Void> redirectIO(final InputStream src, final OutputStream dst) {
        return CompletableFuture.runAsync(() -> {
            try (dst) {
                src.transferTo(dst);
            } catch (Exception e) {
                // silence!
            }
        });
    }

    public static void closeSilent(Closeable handle) {
        try {
            handle.close();
        } catch (Exception e) {
            // silence!
        }
    }

    public static String readAll(final InputStream src) throws IOException {
        ByteArrayOutputStream out = new ByteArrayOutputStream();
        src.transferTo(out);
        return out.toString();
    }

    public static <T> T readJson(final Reader reader, Class<T> type) throws IOException {
        ObjectMapper objectMapper = new ObjectMapper()
                .registerModule(new Jdk8Module());
        return objectMapper.readValue(reader, type);
    }

    public static void writeJson(final File file, Object value)
            throws StreamWriteException, DatabindException, IOException {
        ObjectMapper objectMapper = new ObjectMapper()
                .enable(SerializationFeature.INDENT_OUTPUT)
                .registerModule(new Jdk8Module());
        objectMapper.writeValue(file, value);
    }

    public static ImageRef readImageRef(final File file) throws FileNotFoundException, IOException {
        try (var reader = new FileReader(file)) {
            return readJson(reader, ImageRef.class);
        }
    }

    public static String readAll(final File src) {
        try {
            return src.isFile() ? Files.readString(src.toPath()) : null;
        } catch (IOException e) {
            throw new GradleException("Failed to read file " + src.toString(), e);
        }
    }

    public static List<String> readAll(final Iterable<? extends File> files) {
        return StreamSupport.stream(files.spliterator(), false)
                .map(Utils::readAll)
                .toList();
    }

    public static String readAll(final InputStream src, String charset) throws IOException {
        ByteArrayOutputStream out = new ByteArrayOutputStream();
        src.transferTo(out);
        return out.toString(charset);
    }

    public static JsonGenerator createDefaultJsonGenerator() {
        return new JsonGenerator.Options()
                .excludeNulls()
                .addConverter(new Converter() {
                    public boolean handles(Class<?> type) {
                        return (File.class == type);
                    }

                    public Object convert(Object value, String key) {
                        return ((File) value).getPath();
                    }
                })
                .build();
    }

    public static String toJsonPretty(Object value) {
        return JsonOutput.prettyPrint(createDefaultJsonGenerator().toJson(value));
    }

    public static boolean isNullOrEmptyString(String value) {
        return (value == null || value.length() == 0);
    }

    public static <T> Action<T> wrapClosure(Closure<?> closure) {
        return x -> {
            closure.setDelegate(x);
            closure.setResolveStrategy(Closure.OWNER_FIRST);
            closure.call(x);
        };
    }

    public static Set<String> mapToString(Set<Object> set) {
        return set.stream().map(Object::toString).collect(Collectors.toSet());
    }

    public static File normalizeFile(Object file) {
        if (file == null)
            throw new IllegalArgumentException("file");

        if (file instanceof String)
            return new File((String)file);
        if (file instanceof FileSystemLocation)
            return ((FileSystemLocation)file).getAsFile();
        else if (file instanceof File)
            return (File)file;
        throw new ClassCastException();
    }

}