package org.implab.gradle.containers.tasks;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.gradle.api.Action;
import org.gradle.api.file.Directory;
import org.gradle.api.file.DirectoryProperty;
import org.gradle.api.file.RegularFile;
import org.gradle.api.file.RegularFileProperty;
import org.gradle.api.provider.MapProperty;
import org.gradle.api.provider.Property;
import org.gradle.api.provider.Provider;
import org.gradle.api.tasks.Input;
import org.gradle.api.tasks.InputDirectory;
import org.gradle.api.tasks.Internal;
import org.gradle.api.tasks.Optional;
import org.gradle.api.tasks.OutputFile;
import org.gradle.api.tasks.SkipWhenEmpty;
import org.gradle.api.tasks.TaskAction;
import java.io.File;
import java.io.IOException;

import org.implab.gradle.containers.cli.ImageRef;
import org.implab.gradle.containers.cli.RedirectTo;
import org.implab.gradle.containers.cli.Utils;
import org.implab.gradle.containers.dsl.MapPropertyEntry;
import org.implab.gradle.containers.dsl.OptionsMixin;

import groovy.lang.Closure;

public abstract class BuildImage extends DockerCliTask implements OptionsMixin {

    @InputDirectory
    @SkipWhenEmpty
    public abstract DirectoryProperty getContextDirectory();

    @Input
    public abstract MapProperty<String, String> getBuildArgs();

    @Input
    @Optional
    public abstract Property<String> getBuildTarget();

    @Input
    public abstract Property<Object> getImageName();

    @Internal
    public abstract RegularFileProperty getImageIdFile();

    @OutputFile
    public Provider<File> getImageIdFileOutput() {
        return getImageIdFile().map(RegularFile::getAsFile);
    }

    public BuildImage() {
        getOutputs().upToDateWhen(task -> getImageIdFile()
                .map(this::imageExists)
                .getOrElse(false));
    }

    public void buildArgs(Action<Map<String, String>> spec) {
        getBuildArgs().putAll(provider(() -> {
            Map<String, String> args = new HashMap<>();
            spec.execute(args);
            getLogger().info("add buildArgs {}", args);
            return args;
        }));
    }

    public void buildArgs(Closure<Map<String, String>> spec) {
        buildArgs(Utils.wrapClosure(spec));
    }

    public MapPropertyEntry<String, String> buildArg(String key) {
        return new MapPropertyEntry<String, String>(getBuildArgs(), key, getProviders());
    }

    private boolean imageExists(RegularFile file) {
        return readRefId(file.getAsFile()).map(this::imageExists).orElse(false);
    }

    private java.util.Optional<String> readRefId(File idFile) {
        try {
            return idFile.exists() ? java.util.Optional.of(Utils.readImageRef(idFile).getId())
                    : java.util.Optional.empty();
        } catch (IOException e) {
            getLogger().error("Failed to read imageId {}: {}", idFile, e);
            return java.util.Optional.empty();
        }
    }

    @Override
    protected java.util.Optional<RedirectTo> loggerErrorRedirect() {
        // https://github.com/moby/buildkit/issues/1186
        return loggerInfoRedirect();
    }

    @TaskAction
    public void run() throws Exception {
        List<String> args = new ArrayList<>();

        // create a temp file to store image id
        var iidFile = new File(this.getTemporaryDir(), "imageid");

        // specify where to write image id
        args.addAll(List.of("--iidfile", iidFile.getAbsolutePath()));

        getBuildArgs().get().forEach((k, v) -> {
            args.add("--build-arg");
            args.add(String.format("%s=%s", k, v));
        });

        // add --target if specified for multi-stage build
        if (getBuildTarget().isPresent()) {
            args.add("--target");
            args.add(getBuildTarget().get());
        }

        // add extra parameters
        getOptions().get().forEach(args::add);

        var imageTag = getImageName().map(Object::toString).get();

        // build image
        var spec = docker().imageBuild(
                imageTag,
                getContextDirectory().map(Directory::getAsFile).get(),
                args);

        exec(spec);

        // read the built image id and store image ref metadata
        var imageRef = new ImageRef(imageTag, Utils.readAll(iidFile));
        Utils.writeJson(getImageIdFile().map(RegularFile::getAsFile).get(), imageRef);
    }
}
